import java.awt.image.PixelGrabber;
import java.awt.Image;
import javax.imageio.ImageIO;
import java.io.File;
import java.io.IOException;

public class Bump {
  
  private int[] light;
  private int[] bumpmap;
  private int[] texture;
  
  private int softness = 1;

  private int w;
  private int h;
  
  private int [][] lightColors = 
  {
    {0xff, 0x88, 0},
    {0   , 0xee, 0xff},
    {0x80, 0x80, 0x80}
  };
  
  private int [][][] curves = 
    new int[][][] {
      {{2,40,425}, {-1,40,425}},
      {{3,40,0}, {-2,30,0}, {5, 10, 0}},
      {{-5,40,0}, {3,20,1000}, {-4, 10, 1000}}
    };
  
  private int[][] lxy = {
    {0,0},
    {0,0},
    {0,0}
  };

  private void getImage(String sFile, int[] buffer) {
    try {
      
      Image image = ImageIO.read(getClass().getClassLoader().getResourceAsStream(sFile));
      PixelGrabber pg
        = new PixelGrabber(image,0,0,w,h,buffer,0,w);
      
      pg.grabPixels();
    }
    catch (Exception e) {throw new RuntimeException("Unable to open "+sFile+".");}
  }

  public Bump(int width, int height, int[] buffer1, int[] buffer2) {

    w = width;
    h = height;
    
    bumpmap = buffer1;
    texture = buffer2;
    
    getImage("map.png", bumpmap);
       
    light = new int[256*256];
    
    initLight();
    initBumpmap();

  }

  public void initBumpmap() {
    int offs = 0;
    
    for (int j = 0; j < h; j++) 
      for (int i = 0; i < w; i++, offs++){
	int iOrig = bumpmap[offs];
	int red   = (iOrig>>16) & 0xff;
	int green = (iOrig>>8)  & 0xff;
	int blue  = iOrig       & 0xff;
	
	int I = (int)(red * 0.299 + green * 0.587 + blue * 0.114);//Y in YIQ
	bumpmap[offs] = I;
      }
    softenBumpMap();
  }

  private void softenBumpMap() {
    int [] softMap = new int[w*h];
    int sum;
    for (int s = 0; s < softness; s++) {
      for (int j = 1 ; j < h-1; j++)
	for (int i = 1 ; i < w-1; i++) {
	  sum = 0;
	  sum += bumpmap[(j-1)*w+(i-1)];
	  sum += bumpmap[(j)*w+(i-1)];
	  sum += bumpmap[(j+1)*w+(i-1)];
	  
	  sum += bumpmap[(j-1)*w+(i)];
	  sum += bumpmap[(j)*w+(i)]; 
	  sum += bumpmap[(j+1)*w+(i)];
	  
	  sum += bumpmap[(j-1)*w+(i+1)];
	  sum += bumpmap[(j)*w+(i+1)];
	  sum += bumpmap[(j+1)*w+(i+1)];
	  
	  softMap[j*w+i] = sum/9;
	}
    
      for (int j = 0 ; j < h; j++) {
	softMap[j*w] = bumpmap[j*w];
	softMap[j*w+w-1] = bumpmap[j*w+w-1];
      }
      for (int i = 1 ; i < w-1; i++) {
	softMap[i] = bumpmap[i];
	softMap[(h-1)*w+i] = bumpmap[(h-1)*w+i];
      }
      int n = bumpmap.length;
      for (int i = 0; i < n; i++)
        bumpmap[i] = softMap[i];
    }
  }

  private void initLight() {
    
    for (int i = 0; i < 256; i++)
      for (int j = 0; j < 256;j++) {
	int dx = i-128;
	int dy = j-128;
	int d = dx*dx +dy*dy;
	d = (int)(255-2*Math.sqrt(d));
	light[(i<<8)+j] = (d < 0)? 0: (int)d;
      }
    
    int [] phLight = new int[256];
    int phong = 64;
    for (int i=0;i<(255-phong);i++)
      phLight[i]=(i*i/(255-phong));
    
    for (int i=(255-phong);i<256;i++)
      phLight[i]=(i+(255-i)*(i+phong-255)/phong);
		
    
    for (int i = 0; i < 256*256; i++)
      light[i] = phLight[light[i]]<<1;
  }

  
  private final int getLight(int dx, int dy) {
    if (dx < -127 || dx > 128 || dy < -127 || dy > 128)
      return 0;
    else {
      // System.out.println("i: "+(((128-dy)<<8) + 128-dx));
      return light[((128-dy)<<8) + 128-dx];
    }
  }
  
  
  
  public void render(int [] screen) {
    int offs;
    int max = w*h-1;
    int time = (int)(System.currentTimeMillis());

      
    
    
   
    
    int t = (int)System.currentTimeMillis();
    
    int lights = lightColors.length;
    for (int n = 0; n < lights; n++ ) {
      float fx=0,fy=0;
      for (int i = 0; i < curves[n].length; i++) {
        fx += curves[n][i][1]*Math.cos(curves[n][i][0]*t/(Math.PI*200)
                                   +curves[n][i][2]*Math.PI/1000.0);
        fy += curves[n][i][1]*Math.sin(curves[n][i][0]*t/(Math.PI*200)
                                   +curves[n][i][2]*Math.PI/1000.0);
      }
      
      lxy[n][0] = (w>>1)+(((int)fx)<<1);
      lxy[n][1] = (h>>1)+(int)fy;
    }

  
   
    for (int i = 0; i < w; i++)
      screen[i] = 0xff000000;
    offs = 0;
    for (int i = 1; i < h; i++)
      screen[offs+=w] = 0xff000000;
    offs = w+1;

    for (int j = 1; j < h; j++,offs++) {   
      for (int i = 1; i < w; i++,offs++) {
        int bm = bumpmap[offs];
	int bx = bumpmap[offs-1]-bm;
	int by = bumpmap[offs-w]-bm;

	
	
	int cRed=0;	
	int cGreen=0;
	int cBlue=0;
        
        for (int n = 0; n < lights; n++) {
          int dx = i-lxy[n][0];
          int dy = j-lxy[n][1];
          int c = getLight(bx+dx<<1, by+dy<<1);

          if (c != 0) {
            cRed += c*lightColors[n][0];
            cGreen += c*lightColors[n][1];;
            cBlue += c*lightColors[n][2];;
          }
          
        }

	int iOrig = texture[offs]; 
	int red = (iOrig >> 16) & 0xff;
	red = (red * cRed)>>16; 
	int green = (iOrig >> 8) & 0xff;
	green = (green * cGreen)>>16; 

	int blue = iOrig & 0xff;
	blue = (blue * cBlue)>>16; 

	if (red > 0xff)
	  red = 0xff;
	if (green > 0xff)
	  green = 0xff;
	if (blue > 0xff)
	  blue = 0xff;

	screen[ offs ] = (0xff000000 | (red<<16) | (green<<8) | blue);
	 
	
      }
    }
  }
}
