
import java.util.Random;

public class FireWorks {
    
  private Random r;
  
  private static final int FIRES = 30;
  private static final int NODES = 120;
  private static final int NODE_RADIUS = 2;
  private static final int EXPS = 4;
  
  private static final int FIRELIFE = 20;
  private static final int NODELIFE = 100;
  
  private class FireElement {
    public float x;
    public float y;
    public int life;
    public final void init(float nx, float ny ) {
      double ang = r.nextFloat()*2*Math.PI;
      double ra = r.nextFloat()*NODE_RADIUS;
      x = nx +(int)(ra*Math.cos(ang));
      y = ny +(int)(ra*Math.sin(ang));
      life = r.nextInt(FIRELIFE);
    }
  }
  
  private class FireNode {
    public float x;
    public float y;
    public float dx;
    public float dy;
    public int life;
    public int state;
    
    
    private FireElement[] fires = new FireElement[FIRES];
    
    public static final int STATE_SLEEP  = 0;
    public static final int STATE_FIRE   = 2;
    public static final int STATE_COOLING   = 3;
    
    
    public FireNode() {
      state = STATE_SLEEP;
      for (int i = 0; i < FIRES; i++)
        fires[i] = new FireElement();
    }
    
    public final void init(int fx, int fy, double ang) {
      x = fx;
      y = fy;
      int ra = r.nextInt(200)+30;
      dx = (float)(ra*Math.cos(ang)/100.0);
      dy = (float)(ra*Math.sin(ang)/100.0);
      life = NODELIFE-30+r.nextInt(30);

      for (int i = 0; i < FIRES; i++)
        fires[i].init(x, y);
      
      state = STATE_FIRE;
    }
  
    public final void move() {
      if (life <= 0) {
        if (state == STATE_FIRE) {
          state = STATE_COOLING;
          life = FIRELIFE;
        }
        else {
          state = STATE_SLEEP;
        }
      }
       
      x  += dx;
      y  += dy;
      dx *= 0.98;
      dy *= 0.98;
      
      dy += 0.012;
      life--;
      
      for (int i = 0; i < FIRES; i++) {
        if (fires[i].life <= 0 && state != STATE_COOLING) {
          fires[i].init(x, y);
        }
        else 
          fires[i].life--;
        
      }
    }
    public void heatUp(int [] screen) {
      if (state == STATE_SLEEP)
        return;
          
      int col = 10;
      int [] mask = {
        9,7,6,7,9,
        7,5,3,5,7,
        6,3,0,3,6,
        7,5,3,6,7,
        9,7,6,7,9};
    
      for (int i = 0; i < FIRES; i++) {
        if (fires[i].life <= 0)
          continue;
        int x = (int) fires[i].x;
        int y = (int) fires[i].y;
        if (x < 2 || x >= w -2 || y < 2 || y >=h-2)
          continue;
        int m = 0;
        for (int r = -2; r < 3; r++) {
          for (int c = -2; c < 3; c++) {
            screen[(y+r)*w+c+x] += col-mask[m++];
          }
        }
      }
    }
  }
  
  private class Explosion {
    public int state;
    private int life;
    public static final int STATE_SLEEP  = 0;
    public static final int STATE_FIRE   = 2;
    public FireNode[] nodes = new FireNode[NODES];
    public Explosion() {
      state = STATE_SLEEP;
      for (int i = 0; i < NODES; i++)
        nodes[i] = new FireNode();

    }

    public void update(int [] screen) {
      
      
      if (state == STATE_SLEEP) {
        
        int x = r.nextInt(w);
        int y = r.nextInt(h);
        
        for (int i = 0; i < NODES; i++) {
          int suunta = r.nextInt(NODES);
          double ang = suunta * (2*Math.PI/NODES);
          nodes[i].init(x, y, ang);
        }
        state = STATE_FIRE;
        life = NODELIFE+FIRELIFE+r.nextInt(150);
      }
      else {
        for (int i = 0; i < NODES; i++)
          nodes[i].heatUp(screen);
        
        for (int i = 0; i < NODES; i++)
          if (nodes[i].state != FireNode.STATE_SLEEP)
            nodes[i].move();
        
        life--;
        if (life <= 0) {
          state = STATE_SLEEP;
        }
      }
    }
  }


  private Explosion[] exps = new Explosion[EXPS];
  
  private int[] colors = new int[256];

  private int w, h;

  public FireWorks(int width, int height) {
    r = new Random(0xcafebabe);

    w = width;
    h = height;

    for (int i = 0; i < EXPS; i++)
      exps[i] = new Explosion();
  }

  private void rotateColors() {

    long aika = System.currentTimeMillis();

    for (int i = 0; i < 32; i++) {
      colors[i] = 0xff000000 | (i<<16) | (i<<8) | i;
    }
    // musta -> punainen

    int r = (int)(72 + 71 * Math.cos( aika / 740.0));
    int g = (int)(72 + 71 * Math.sin( aika / 630.0));
    int b = (int)(72 - 71 * Math.cos( aika / 810.0));
    r &= 0xff;
    g &= 0xff;
    b &= 0xff;

    for (int i = 32; i < 128; i++) {
      int rr = 32+((int)(( (i-32) * (r-32) )/96.0))<<16;
      int gg = 32+((int)(( (i-32) * (g-32) )/96.0))<<8;
      int bb = 32+((int)(( (i-32) * (b-32) )/96.0));
      colors[i] = 0xff000000 | rr | gg | bb ;
    }    
    
    int r2 = (int)(142 + 71 * Math.cos( aika / 340.0));
    int g2 = (int)(142 + 71 * Math.sin( aika / 430.0));
    int b2 = (int)(142 - 71 * Math.cos( aika / 910.0));
    
    r2 &= 0xff;
    g2 &= 0xff;
    b2 &= 0xff;
    
    
    for (int i = 128; i < 192; i++) {

      int rr = r+((int)(( (i-128) * (r2-r) )/64.0))<<16;
      int gg = g+((int)(( (i-128) * (g2-g) )/64.0))<<8;
      int bb = b+((int)(( (i-128) * (b2-b) )/64.0));
      colors[i] = 0xff000000 | rr | gg | bb ;
    }
    
    r = 255;
    g = 255;
    b = 255;
    
    for (int i = 192; i < 256; i++) {
      int rr = r2+((int)(( (i-192) * (r-r2) )/64.0))<<16;
      int gg = g2+((int)(( (i-192) * (g-g2) )/64.0))<<8;
      int bb = b2+((int)(( (i-192) * (b-b2) )/64.0));
      colors[i] = 0xff000000 | rr | gg | bb ;
    }

  }

 
  public final void render(int [] screen) {
    int n = w*h;
    rotateColors();

    for (int i = 0; i < n; i++) 
      screen[i] = 0;
   
    for (int i = 0; i < EXPS; i++)
      exps[i].update(screen);
    
    for (int i = 0; i < n; i++) {
      int I = screen[i]<< 1;
      screen[i] = (I >= 256) ? 0xffffffff : colors[I];
    }
  }

}
