import java.util.Random;

public class Mato {
    
  private int[] buf;

  private Random r;
  
  private static final int MATOS = 15;
  
  private float[] matoX    = new float[MATOS];
  private float[] matoY    = new float[MATOS];
  private float[] matoSpeedX = new float[MATOS];
  private float[] matoSpeedY = new float[MATOS];
  private float[] matoAcc = new float[MATOS];
  
  private int[] matoColor  = new int[MATOS];
  
  private int[] colors = new int[256];

  private int w; 
  private int h;

  public Mato(int width, int height, int[] buffer) {

    w = width;
    h = height;
    r = new Random(0xF1F0);
    buf = buffer;

    initMatos();
    initColors();
  }

  private static final double sq(double d) { return d*d;};
  private void initColors() {
    colors[0] = 0xff000000;
    for (int i = 1; i < 256; i++) {
      int r = 0xff & (int)(255*sq(Math.sin((i<<4)*Math.PI/256)));
      int g = 0xff & (int)(255*sq(Math.sin((i<<3)*Math.PI/256+1.212)));
      int b = 0xff & (int)(255*sq(Math.sin((i<<4)*Math.PI/256+2.234)));
      colors[i] = 0xff000000 | (r << 16) | (g << 8) | b;
    }
  }
  
  private void rotateColors() {
    int col = colors[1];
     for (int i = 1; i < 255; i++)
       colors[i] = colors[i+1];
     colors[255] = col;
  }

  private void initMatos() {
    for (int i = 0 ; i < MATOS; i++) {
      matoX[i] = w>>1;
      matoY[i] = h>>1;
      matoSpeedX[i] = 2*r.nextFloat();
      matoSpeedY[i] = 2*r.nextFloat();

      matoAcc[i] = (float)(Math.PI*2*r.nextFloat());
      matoColor[i] = r.nextInt(255);
    }
    
  }

  public final void render(int [] background, int[] screen) {
    int n = w*h;
    rotateColors();


    int [] mask = {
      0,0,1,1,1,1,0,0,
      0,1,1,1,1,1,1,0,
      1,1,1,1,1,1,1,1,
      1,1,1,1,1,1,1,1,
      1,1,1,1,1,1,1,1,
      1,1,1,1,1,1,1,1,
      0,1,1,1,1,1,1,0,
      0,0,1,1,1,1,0,0
    };

    for (int i = 0; i < MATOS; i++) {
      int x = (int) matoX[i];
      int y = (int) matoY[i];

      matoAcc[i] += 0.4*(r.nextFloat()-0.5);
      double sx = matoSpeedX[i];
      double sy = matoSpeedY[i];
      sx+= Math.cos(matoAcc[i]);
      sy+= Math.sin(matoAcc[i]);
      double d = Math.sqrt(sx*sx+sy*sy);
     
      matoSpeedX[i] = (float) (2*sx / d);
      matoSpeedY[i] = (float) (2*sy / d);
      
      matoX[i] += matoSpeedX[i];
      matoY[i] += matoSpeedY[i];
      
      if (matoX[i] < 0) matoX[i] += w;
      else if (matoX[i] >= w) matoX[i] -= w;
      if (matoY[i] < 0) matoY[i] += h;
      else if (matoY[i] >= h) matoY[i] -= h;

      matoColor[i]++;

      if (matoColor[i] >= 255) matoColor[i] = 1;

      int m = 0;
      for (int r = -4; r < 4; r++) {
        for (int c = -4; c < 4; c++) {
          
          if (mask[m++] != 0) {
            int yr = y+r;
            if (yr < 0 || yr >= h)
              continue;
            int xc = x+c;
            if (xc < 0 || xc >= w)
              continue;
            buf[yr*w+xc] = matoColor[i];
          }
        }
      }

    }
    
    for (int i = 0; i < n; i++) {
      int b = buf[i];
      if (b > 0)
        screen[i] = colors[buf[i]];
      else
        screen[i] = background[i];
    }
  }
}
