public class Plasma  {
  private byte[][] buffer1;
  private byte[][] buffer2;


  private int h,w;
  private int s;
  private int aika;  

  private int [] paletti;
  public Plasma(int width, int height) {
    w = width;
    h = height;
    aika = 3000;
    s = w > h ? w : h;
    
    buffer1=new byte[3*s][3*s];
    buffer2=new byte[3*s][3*s];
    paletti = new int[256];  

    initBuffers();
  }
  
  private void initBuffers() {
    int H = 3*s;
    int W = 3*s;
    double [] sins = new double[256];
    double [] coss = new double[256];
    for (int i = 0; i < 256; i++) {
      sins[i] = Math.sin(i*Math.PI/128.0);
      coss[i] = Math.cos(i*Math.PI/128.0);
    }
    for(int i = 0; i < W; i++) {
      int ang = 0xff & (int)((i<<8) / (60.0*2*Math.PI));
      double a = sins[ang];
      for (int j = 0; j < H; j++) {
        ang = 0xff & (int)((j<<8) / (74.0*2*Math.PI));
        double b = coss[ang];
        ang = 0xff & (int)((j<<8) / (2*Math.PI*(30.0 + 5.0 * a)));
        double c = coss[ang];
        ang = 0xff & (int)((i<<8) / (2*Math.PI*(40.0 + 10 * b)));
        double d = sins[ang];
        buffer1[j][i] = (byte) (64.0 + 63.0 * d * c);
      }
    }
    for(int i = 0; i < W; i++) {
      int ang = 0xff & (int)((i<<8) / (50.0*2*Math.PI));
      double a = sins[ang];
      ang = 0xff & (int)((i<<8) / (-45.353*2*Math.PI));
      double b = sins[ang];
      for (int j = 0; j < H; j++) {
        ang = 0xff & (int)((j<<8) / (2*Math.PI*(64.2 + 3.0 * a)));
        double c = coss[ang];
        buffer2[j][i] = (byte)(64.0 + 63.0 * b * c);
      
      }
    }
  }

  public boolean render(int [] screen) {
    
    aika++;
    int max = w*h;
    

    for(int i = 0; i < 256; i++) {   
      int red   = (int)(72 + 71 * Math.cos(i * Math.PI / 128 + aika / 74.0));
      int green = (int)(72 + 71 * Math.sin(i * Math.PI / 128 + aika / 63.0));
      int blue  = (int)(72 - 71 * Math.cos(i * Math.PI / 128 + aika / 81.0));
      paletti[i] = 0xff000000 | (red << 16) | (green << 8) | blue;
    }
    int r = s/6;


    int bx1 = (int)(3*s/2 + r*Math.cos(4.357*aika*Math.PI/360));
    int by1 = (int)(3*s/2 + r*Math.sin(4.357*aika*Math.PI/360));
    int bx2 = (int)(3*s/2 + r*Math.sin(-1.7*aika*Math.PI/360));
    int by2 = (int)(3*s/2 + r*Math.cos(-1.7*aika*Math.PI/360));
    double cs = Math.cos(aika*Math.PI/360);
    double ss = Math.sin(aika*Math.PI/360);
    
    int offs = 0;
    for (int y = 0; y < h; y++)
      for (int x = 0; x < w; x++) {
        int arvo1 = buffer1[by1+(int)(y*ss)][bx1+(int)(x*cs)];
        int arvo2 = buffer2[by2+(int)(y*cs)][bx2+(int)(x*ss)];
        
        screen[offs++] = paletti[arvo1+arvo2];
      }
    
    return true;
  }
}
