import java.util.Arrays;
import java.util.Comparator;
import java.util.Vector;
public class Slo3d {
  
  private int[] texture;

  private int[] rowmin;
  private int[] rowmax;
  private double[] texminx;
  private double[] texminy;
  private double[] texmaxx;
  private double[] texmaxy;
  

  private boolean bPlotTexture = true;
  private boolean bPlotWireframe = false;
  private boolean bPlotVertex = false;
  
  private int w, h;

  
  private class Vertex {
    public float x,y,z;
    public float px, py, pz;
    //texture mapping
    public int tx,ty;
    public int color;
    
  }
  
  private class Quad {
    public int[] v;
    public Quad() {
      v = new int[4];
    }
  }

  private static class Matrix {
    public static double[][] getIdentity() {
      double [][] m = new double[4][4];
      for (int i = 0; i < 4; i++)
        m[i][i] = 1;
      return m;
    }
    public static double[][] getRotX(double ang) {
      double [][] m = new double[4][4];
      
      double c = Math.cos(ang);
      double s = Math.sin(ang);
      m[0][0] = 1;
      m[1][1] = c;
      m[1][2] = s;
      m[2][1] = -s;
      m[2][2] = c;
      m[3][3] = 1;
      return m;
    }
    
    public static double[][] scale(double s) {
      double [][] m = new double[4][4];
      m[0][0] = s;
      m[1][1] = s;
      m[2][2] = s;
      m[3][3] = 1;
      return m;
    }
    
    public static double[][] getRotY(double ang) {
      double [][] m = new double[4][4];
      
      double c = Math.cos(ang);
      double s = Math.sin(ang);
      m[0][0] = c;
      m[0][2] = -s;
      m[1][1] = 1;
      m[2][0] = s;
      m[2][2] = c;
      m[3][3] = 1;
      return m;
    }
    public static double[][] getRotZ(double ang) {
      double [][] m = new double[4][4];
      
      double c = Math.cos(ang);
      double s = Math.sin(ang);
      m[0][0] = c;
      m[0][1] = s;
      m[1][0] = -s;
      m[1][1] = c;
      m[2][2] = 1;
      m[3][3] = 1;
      return m;
    }
    
    public static double[][] mult(double[][] m1, double[][] m2) {
      double [][] m = new double[4][4];
      for(int i = 0; i < 4; i++)
        for(int j = 0; j < 4; j++)
          for(int k = 0; k < 4; k++)
            m[i][j] += m1[i][k]*m2[k][j];

      return m;
    }
  }

  private class Object3d {
    private int origox;
    private int origoy;
    
    public Vertex[] vs;
    public Quad[] qs;
    public double[][] transform;
    public Object3d(int x, int y) {
      origox = x;
      origoy = y;
      transform = Matrix.getIdentity();
    }
    
    public void plot(int[] screen) {
      int n = vs.length;
      int max = w*h;

      for (int i = 0; i < n; i++) {
        double x1 = 
          transform[0][0] * vs[i].x
          + transform[0][1] * vs[i].y
          + transform[0][2] * vs[i].z
          + transform[0][3];
        double y1 = 
          transform[1][0] * vs[i].x
          + transform[1][1] * vs[i].y
          + transform[1][2] * vs[i].z
          + transform[1][3];
        
        double z = 
          transform[2][0] * vs[i].x
          + transform[2][1] * vs[i].y
          + transform[2][2] * vs[i].z
          + transform[2][3];

        double zoom = 1+ z / 1000.0;
        if (zoom < 0)
          zoom = 0;

        x1 *= zoom;
        y1 *= zoom;
        vs[i].px = (float)x1;
        vs[i].py = (float)y1;
        vs[i].pz = (float)z;
        
        if (bPlotVertex) {
          int index = origox+((int)x1)+w*(origoy+((int)y1));
          if (index >= 0 && index < max)
            screen[index] = vs[i].color; 
        }
      }
        
      if (bPlotWireframe) {
        for (int i = 0; i < qs.length; i++) {
          Quad q = qs[i];
          plotLine(vs[q.v[0]].px, vs[q.v[0]].py,vs[q.v[1]].px, vs[q.v[1]].py, screen);
          plotLine(vs[q.v[0]].px, vs[q.v[0]].py,vs[q.v[2]].px, vs[q.v[2]].py, screen);
          plotLine(vs[q.v[2]].px, vs[q.v[2]].py,vs[q.v[3]].px, vs[q.v[3]].py, screen);
          plotLine(vs[q.v[1]].px, vs[q.v[1]].py,vs[q.v[3]].px, vs[q.v[3]].py, screen);
        }
      }     
      
      if (bPlotTexture) {
        Arrays.sort(qs, new Comparator() {
            public int compare(Object o1, Object o2) {
              Quad q1 = (Quad)o1;
              Quad q2 = (Quad)o2;
              
              if (vs[q1.v[0]].pz+vs[q1.v[3]].pz < vs[q2.v[0]].pz+vs[q2.v[3]].pz )
                return -1;
              else
                return 1;
            }
          });
     
        for (int i = 0; i < qs.length; i++) {
          Quad q = qs[i];
          for (int j = 0; j < h; j++){
            rowmin[j] = 320;
            rowmax[j] = -1;
            texminx[j] = 0;
            texminy[j] = 0;
            texmaxx[j] = 0;
            texmaxy[j] = 0;
          }
          plotTextureLine(vs[q.v[0]].px, vs[q.v[0]].py, vs[q.v[1]].px, vs[q.v[1]].py, 
                          vs[q.v[0]].tx, vs[q.v[0]].ty, vs[q.v[1]].tx, vs[q.v[1]].ty, 
                          screen,texture);
          plotTextureLine(vs[q.v[0]].px, vs[q.v[0]].py, vs[q.v[2]].px, vs[q.v[2]].py, 
                          vs[q.v[0]].tx, vs[q.v[0]].ty, vs[q.v[2]].tx, vs[q.v[2]].ty, 
                          screen,texture);
          
          plotTextureLine(vs[q.v[1]].px, vs[q.v[1]].py, vs[q.v[3]].px, vs[q.v[3]].py, 
                          vs[q.v[1]].tx, vs[q.v[1]].ty, vs[q.v[3]].tx, vs[q.v[3]].ty, 
                          screen,texture);
          
          plotTextureLine(vs[q.v[2]].px, vs[q.v[2]].py, vs[q.v[3]].px, vs[q.v[3]].py, 
                          vs[q.v[2]].tx, vs[q.v[2]].ty, vs[q.v[3]].tx, vs[q.v[3]].ty, 
                          screen,texture);
          
          for (int j = 0; j < h; j++) {
            if (rowmax[j] != -1) {
              
              plotScanLine(rowmin[j], rowmax[j],j, 
                           texminx[j], texminy[j], texmaxx[j], texmaxy[j],
                           screen,texture);
              
            }
          }
          
        }
      }
    }
    
    private void plotLine(float x1, float y1, float x2, float y2, int[] screen) {
      double dx =x2-x1;
      double dy =y2-y1;
      double d = Math.sqrt(dx*dx+dy*dy);
      int n = ((int)d)+1;
      dx /= n;
      dy /= n;
      for (int i = 0; i < n; i++,x1+=dx, y1+=dy) {
        int x = origox +(int)x1;
        int y = origoy +(int)y1;
        if (x >= 0 && x < w && y >= 0 && y < h)
          screen[x+w*y] = 0xffffffff;
      }
    }
    private void plotScanLine(int x1, int x2, int y,
                              double tx1, double ty1, double tx2, double ty2,
                              int[] screen, int[] texture) {
      
      
      int n = x2-x1-1;
      int yw = y*w;
      double tdx =tx2-tx1;
      double tdy =ty2-ty1;
      tdx /= n;
      tdy /= n;

      x1 += yw;
      x2 += yw;
       
      for (int x = x1+1; x < x2; x++, tx1+=tdx, ty1+=tdy) {
        int tx = (int)tx1;
        int ty = (int)ty1;
        if (tx < 0)
          tx = 0;
        else if (tx >= texw)
          tx = texw-1;
        if (ty < 0)
          ty = 0;
        else if (ty >= texh)
          ty = texh-1;
        
        screen[x] = texture[tx+texw*ty];
      }
    }
    
    private void plotTextureLine(float x1, float y1, float x2, float y2,
                                 float tx1, float ty1, float tx2, float ty2,
                                 int[] screen, int[] texture) {
      double dx =x2-x1;
      double dy =y2-y1;
      double d = Math.sqrt(dx*dx+dy*dy);
      double tdx =tx2-tx1;
      double tdy =ty2-ty1;
      
      
      int n = ((int)d)+1;
      dx /= n;
      dy /= n;
      tdx /= n;
      tdy /= n;
      n++;
      for (int i = 0; i < n; i++, x1+=dx, y1+=dy, tx1+=tdx, ty1+=tdy) {
        int x = origox +(int)x1;
        int y = origoy +(int)y1;
        if (x >= 0 && x < w && y >= 0 && y < h) {
          if (rowmin[y] > x) { 
            rowmin[y] = x;
            texminy[y] = ty1;
            texminx[y] = tx1;
          }
          if (rowmax[y] < x) {
            rowmax[y] = x;
            texmaxy[y] = ty1;
            texmaxx[y] = tx1;
          }
          
          int tx = (int)tx1;
          int ty = (int)ty1;
          
          if (tx < 0)
            tx = 0;
          else if (tx >= texw)
            tx = texw-1;
          if (ty < 0)
            ty = 0;
          else if (ty >= texh)
            ty = texh-1;
          
          screen[x+w*y] = texture[tx + texw*ty];
          
        }
      }
    }
    
  }

  
  private Object3d o3d;

  private Object3d makeYear() {
    int[][] num2 = {
      {1,16}, {8,13}, {16,11}, {19,12}, {24,15}, {24,19},
      {24,24}, {19,32}, {9,43}, {0,49},  {0,59}, {18,59}, {41,59},
      
      {41,49}, {30,49},{18,49}, {18,48}, {27,41}, {35,32}, {39,22},
      {39,15}, {36,7}, {28,2}, {19,0}, {8,1}, {1,3}
    };

    int[][] num0 = {
      {69,0}, {59,2}, {54,6}, {49,14}, {47,24}, {47,38},
      {49,46}, {53,53}, {60,59}, {68,60},  {75,60},  {82,57},
      {87,52}, {91,43},{92,35},{92,25},{91,15},{88,10},
      {83,4},{78,1},{72,0},

      {69,11}, {66,12}, {64,15}, {63,18}, {62,26}, {62,37},
      {63,43}, {65,48}, {67,50}, {69,50}, {72,49}, {74,48},
      {75,46}, {76,42},{77,34},{77,26},{76,19},{75,15},
      {74,12},{73,12},{71,11}
    };
    
    int[][] num5 = {
      {191,1}, {153,1}, {153,34}, {160,32}, {167,32}, {173,33},
      {176,36}, {177,37}, {178,39}, {178,43},  {177,45},  {175,47},
      {174,48}, {168,49}, {164, 49},{157,48}, {150,44},
      
      {191,11}, {166,11}, {166,21}, {169,22}, {176,22}, {180,23},
      {188,28}, {191,33}, {192,38}, {192,44}, {190,51}, {185,56},
      {179,59}, {172, 60}, {162,60}, {155,58},{150,57}
    };
    int cx = 94;
    int cy = 31;
    
    Object3d o3d = new Object3d(w>>1, h>>1);
    Vertex[] vs = new Vertex[2*num2.length+4*num0.length+2*num5.length];
    int idx = 0;
    for (int i = 0; i < num2.length; i++) {
      vs[idx] = new Vertex();
      vs[idx].x = -cx+num2[i][0];
      vs[idx].y = -cy+num2[i][1];
      vs[idx].z = 10;
      vs[idx].tx = 10+num2[i][0];
      vs[idx].ty = 10+num2[i][1];

      vs[idx++].color = 0xffffffff;
    }
    for (int i = 0; i < num2.length; i++) {
      vs[idx] = new Vertex();
      vs[idx].x = -cx+num2[i][0];
      vs[idx].y = -cy+num2[i][1];
      vs[idx].z = -10;
      vs[idx].tx = 10+num2[i][0];
      vs[idx].ty = 10+num2[i][1];

      vs[idx++].color = 0xffffffff;
    }
    Vector qs = new Vector();
    int n = num2.length;
    for (int i = 0; i < (n>>1) - 1; i++) {
      Quad q = new Quad();
      q.v[0] = n-1-i;
      q.v[1] = i;
      q.v[2] = n-2-i;
      q.v[3] = i+1;
      qs.addElement(q);

      q = new Quad();
      q.v[0] = n+n-1-i;
      q.v[1] = n+i;
      q.v[2] = n+n-2-i;
      q.v[3] = n+i+1;
      qs.addElement(q);
    }
    for (int i = 0; i < n; i++) {
      Quad q = new Quad();
      q.v[0] = i;
      q.v[2] = (i+1)%n;
      q.v[1] = n+i;
      q.v[3] = n+(i+1)%n;
      qs.addElement(q);
    }

    int xplus = 0;
    int n2 = num2.length << 1;
    for (int j = 0; j < 2; j++,xplus +=50, n2+=num0.length<<1) {
      n = num0.length;
      for (int i = 0; i < n; i++) {
        vs[idx] = new Vertex();
        vs[idx].x = xplus-cx+num0[i][0];
        vs[idx].y = -cy+num0[i][1];
        vs[idx].z = 10;
        vs[idx].tx = xplus+10+num0[i][0];
        vs[idx].ty = 10+num0[i][1];
        
        vs[idx++].color = 0xffffffff;
      }
      for (int i = 0; i < n; i++) {
        vs[idx] = new Vertex();
        vs[idx].x = xplus-cx+num0[i][0];
        vs[idx].y = -cy+num0[i][1];
        vs[idx].z = -10;
        vs[idx].tx = xplus+10+num0[i][0];
        vs[idx].ty = 10+num0[i][1];
        
        vs[idx++].color = 0xffffffff;
        
      }
      n = n>>1;
      
      for (int i = 0; i < n; i++) {
        Quad q = new Quad();
        q.v[0] = n2 + n + i;
        q.v[1] = n2 + i;
        q.v[2] = n2 + n + (i+1)%n;
        q.v[3] = n2 + (i+1)%n;
        qs.addElement(q);
        
        q = new Quad();
        q.v[0] = n2 + (n<<1) + n + i;
        q.v[1] = n2 + (n<<1) + i;
        q.v[2] = n2 + (n<<1) + n + (i+1)%n;
        q.v[3] = n2 + (n<<1) + (i+1)%n;
        qs.addElement(q);
        
        q = new Quad();
        q.v[0] = n2          + n + i;
        q.v[1] = n2 + (n<<1) + n + i;
        q.v[2] = n2 +        + n + (i+1)%n;
        q.v[3] = n2 + (n<<1) + n + (i+1)%n;
        qs.addElement(q);
      
        q = new Quad();
        q.v[0] = n2          +  i;
        q.v[1] = n2 + (n<<1) +  i;
        q.v[2] = n2 +        +  (i+1)%n;
        q.v[3] = n2 + (n<<1) +  (i+1)%n;
        qs.addElement(q);
      }
    }
    n = num5.length;
    for (int i = 0; i < n; i++) {
      vs[idx] = new Vertex();
      vs[idx].x = -cx+num5[i][0];
      vs[idx].y = -cy+num5[i][1];
      vs[idx].z = 10;
      vs[idx].tx = 10+num5[i][0];
      vs[idx].ty = 10+num5[i][1];
      
      vs[idx++].color = 0xffffffff;
    }
    for (int i = 0; i < n; i++) {
      vs[idx] = new Vertex();
      vs[idx].x = -cx+num5[i][0];
      vs[idx].y = -cy+num5[i][1];
      vs[idx].z = -10;
      vs[idx].tx = 10+num5[i][0];
      vs[idx].ty = 10+num5[i][1];
      
      vs[idx++].color = 0xffffffff;
    }
    n2 = num2.length*2 + 4*num0.length;
    int nh = n>>1;
    
    for (int i = 0; i < nh-1; i++) {
      Quad q = new Quad();
      q.v[0] = n2 + nh + i;
      q.v[1] = n2 + i;
      q.v[2] = n2 + nh + (i+1)%nh;
      q.v[3] = n2 + (i+1)%nh;
      qs.addElement(q);
      
      q = new Quad();
      q.v[0] = n2 + n + nh + i;
      q.v[1] = n2 + n + i;
      q.v[2] = n2 + n + nh + (i+1);
      q.v[3] = n2 + n + (i+1);
      qs.addElement(q);
    } 
    
    for (int i = 0; i < nh-1; i++) {
      Quad q = new Quad();
      q.v[0] = n2     + i;
      q.v[1] = n2 + n + i;
      q.v[2] = n2     + (i+1);
      q.v[3] = n2 + n + (i+1);
      
      qs.addElement(q);
      q = new Quad();
      q.v[0] = n2+nh     + i;
      q.v[1] = n2+nh + n + i;
      q.v[2] = n2+nh     + (i+1);
      q.v[3] = n2+nh + n + (i+1);
      
      qs.addElement(q);
      
    }
    
    {
      Quad q = new Quad();
      q.v[0] = n2 +nh;
      q.v[1] = n2 + n + nh;
      q.v[2] = n2    ;
      q.v[3] = n2 + n;
      
      qs.addElement(q);
      
      q = new Quad();
      q.v[0] = n2 + n-1  -nh;
      q.v[1] = n2 + n+n-1 -nh;
      q.v[2] = n2 + n-1;
      q.v[3] = n2 + n+n-1;          
      qs.addElement(q);
    }
    o3d.vs = vs;
    o3d.qs = (Quad[])qs.toArray(new Quad[0]);
    return o3d;
  }

  private Plasma plasma;
  
  private int texw;
  private int texh;
  
  public Slo3d( int width, int height ) {
    w = width;
    h = height;
    
    rowmin = new int[h];
    rowmax = new int[h];
    texminx = new double[h];
    texminy = new double[h];
    texmaxx = new double[h];
    texmaxy = new double[h];

    texw = 200;
    texh = 200;
    texture =  new int[texw*texh];
    

    plasma = new Plasma(texw, texh);
    plasma.render(texture);

    o3d = makeYear();
      
  }
  
  private double [][] last = Matrix.getIdentity();//Matrix.scale(50);
  
  private double angx = 0.0239;
  private double angy = 0.017;
  private double angz = -0.013;
  private double dangx = 0.000053;
  private double dangy = 0.000217;
  private double dangz = 0.00013;
  
  public void render(int [] screen) {
    int max = w*h;
    plasma.render( texture );
    for (int i = 0; i < max; i++)
      screen[i] = 0xff000000; 

    angx += dangx;
    if (angx > 0.03 || angx < -0.03)
      dangx = -dangx;
    angy += dangy;
    if (angy > 0.03 || angy < -0.03)
      dangy = -dangy;
    angz += dangz;
    if (angz > 0.03 || angz < -0.03)
      dangz = -dangz;
      

    double [][] m = Matrix.mult(Matrix.getRotY(angy),Matrix.getRotX(angx) );
    m = Matrix.mult(Matrix.getRotZ(angz), m);
    o3d.transform = Matrix.mult(m, last);
    last = o3d.transform;
    
    o3d.plot(screen);
    

  }

}
