
public class Tunnel {
  
  private int[] texture;

  private int[] atan;
  private int[] sqrt;
  
  private int h,w;
    
  

  public Tunnel(int width, int height, int [] buffer1, int [] texture) {
    w = width;
    h = height;
  
    atan = buffer1;
    sqrt = new int[w*h];
    makeTables();
    this.texture = texture;
    fadeSides();
  }
  
  private void makeTables() {
    int miny = -(h>>1);
    int maxy = h-(h>>1);
    int minx = -(w>>1);
    int maxx = w-(w>>1);
    int dy2 = miny*miny;
    
    int offs = 0;
    for (int dy = miny; dy < maxy; dy2 += (dy++<<1)+1) {
      for (int dx = minx, dx2 = minx*minx; dx < maxx; dx2 += (dx++<<1)+1) {
        
        atan[offs]   = (int)((Math.atan2(dy,dx)/Math.PI+1)*w);
        sqrt[offs++] = (int)Math.sqrt(dx2+dy2); 
      }
    }
  }

  private void fadeSides() {
    for (int y = 0; y < h; y++) {
      int c1 = texture[y*w+(w-10)];
      int c2 = texture[y*w+10];
      int r1 = (c1>>16) & 0xff;
      int r2 = (c2>>16) & 0xff;
      int g1 = (c1>>8)  & 0xff;
      int g2 = (c2>>8)  & 0xff;
      int b1 =  c1      & 0xff;
      int b2 =  c1      & 0xff;
      
      double dr = (r2-r1)/20.0;
      double dg = (g2-g1)/20.0;
      double db = (b2-b1)/20.0;
      for (int i = 0; i < 20; i++) {
        int r = (int)(r1+i*dr);
        int g = (int)(g1+i*dg);
        int b = (int)(b1+i*db);
        
        texture[y*w+(w-10+i)%w] = 0xff000000 | (r<<16) | (g<<8) | b; 
        
      }
    }
  }

  private int rot = 0;
  private int radd = 0;
  private int rd = 1;
  private int move = 0;
  
  public void render(int [] screen) {
   
    int holesize = 7;
    int h2 = h << 1;

    rot += radd>>4;
    if (rot >= w)
      rot -=w;
    else if (rot < 0)
      rot += w;
      
    radd += rd;
    if (radd > 256 || radd < -256)
      rd = -rd;
    move -= 7;
    
    if (move < 0)
      move += h2;
    
    int max = w*h;
    
    for (int offs = 0; offs < max; offs++ ) {
      
      int ix = atan[offs];
      int iy = sqrt[offs];
      
      if (iy < holesize)
        screen[offs] = 0xff000000;
      else {
        ix += rot;
        if (ix < 0) {
          do{ ix += w; } while (ix < 0);
        }
        else if (ix >=w ) {
          do{ ix -= w; } while (ix >= w);
        }
        
        iy = iy-holesize+move;
        if (iy < 0) {
          do{ iy += h2; } while (iy < 0);
        }
        else if (iy >= h2 ) {
          do{ iy -= h2; } while (iy >= h2);
        }
        
        if (iy >= h)
          iy = h2-iy-1;
        
        screen[offs] = texture[iy*w+ix];
        
      }
    }    
  }
}
